SystemOrganization addCategory: #'LiCoR-Predicates'!

LogicStorage subclass: #SmalltalkReasoning
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LiCoR-Predicates'!

SmalltalkReasoning subclass: #BasicReasoning
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LiCoR-Predicates'!

!BasicReasoning commentStamp: '<historical>' prior: 0!
This layer contains some rather primitive predicates for reasoning about source code entities and their relations. However, they are not primitive in the sense that they can be written in terms of the predicates provided by the underlying layers of Reification and ParsetreeTraversal. For this reason we define them in a separate layer.!

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
areArgumentsOfMethodXM2

"Retrieve the list of ?Arguments of a ?Method"

?Arguments areArgumentsOfMethod: ?Method if 
	?Method isMethod,RBMethodNode(?,?Arguments,?) isParseTreeOfMethod: ?Method! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
areStatementsOfMethodXM2

"Retrieve the ?Statements of a ?Method"

?Statements areStatementsOfMethod: ?Method if 
	?Method isMethod,RBMethodNode(?,?,RBSequenceNode(?,?Statements)) isParseTreeOfMethod: ?Method! !

!BasicReasoning methodsFor: 'classes' stamp: 'sp 24/10/11 15:26'!
classCallsMethodXM2

"Find all methods ?Method that potentially get called from within ?Class"

?Class classCallsMethod: ?Method if 
	?Caller isMethodInClass: ?Class,?Caller methodCallsMethod: ?Method! !

!BasicReasoning methodsFor: 'typing' stamp: 'sp 24/10/11 15:26'!
classHasInstVarXofTypeXM3

"if classWithInstvarOfType([Point], x, ?T)"

?Class classHasInstVar: ?instVarName ofType: ?Type if 
	?instVarName isInstanceVariableInClass: ?Class,[MLI forSmalltalk instanceVariableTypesOf: ?instVarName inClass: ?Class] contains: ?Type! !

!BasicReasoning methodsFor: 'classes' stamp: 'sp 24/10/11 15:26'!
classImplementsSelectorXM2

"Does ?Class implement a method with selector ?Selector"

?Class classImplementsSelector: ?Selector if 
	?Class isClass,? isMethodWithName: ?Selector inClass: ?Class! !

!BasicReasoning methodsFor: 'classes' stamp: 'sp 24/10/11 15:26'!
classInheritsFromXM2

"The class ?Sub belongs to the class hierarchy below ?Super, but excluding the class ?Super itself."

?Sub classInheritsFrom: ?Super if 
	?Sub isClassInHierarchyOf: ?Super,not(?Sub equals: ?Super)! !

!BasicReasoning methodsFor: 'naming' stamp: 'sp 24/10/11 15:26'!
classMethodWithNameInClassM3

"defines the name (selector) of a CLASS method in a class."

classMethodWithNameInClass(?method,?selector,?class) if 
	classMethodInClass(?method,?class),methodWithName(?method,?selector)! !

!BasicReasoning methodsFor: 'classes' stamp: 'sp 24/10/11 15:26'!
classUnderstandsSelectorXM2

"Class ?Class understands a message with selector ?Selector.
So the class implements the selector itself, or it is implemented by one of its superclasses.

Case 1: ?Selector is bound"

?Class classUnderstandsSelector: +?Selector if 
	?Class isClass,[MLI forSmalltalk class: ?Class understands: ?Selector].
"Class ?Class understands a message with selector ?Selector.
So the class implements the selector itself, or it is implemented by one of its superclasses.

Case 2: ?Selector is not bound"

?Class classUnderstandsSelector: -?Selector if 
	?Class isClass,[MLI forSmalltalk allSelectorsUnderstoodByClass: ?Class] contains: ?Selector! !

!BasicReasoning methodsFor: 'naming' stamp: 'sp 24/10/11 15:26'!
classVariableWithNameM2

"defines the name of a class variable"

classVariableWithName(?ClassVariable,?name) if 
	classVariable(?ClassVariable),equals(?name,[MLI current classVariableNameOf: ?ClassVariable])! !

!BasicReasoning methodsFor: 'naming' stamp: 'sp 24/10/11 15:26'!
instanceVariableWithNameM2

"defines the name of an instance variable"

instanceVariableWithName(?instanceVariable,?name) if 
	instanceVariable(?instanceVariable),equals(?name,[MLI current instanceVariableNameOf: ?instanceVariable])! !

!BasicReasoning methodsFor: 'classes' stamp: 'sp 24/10/11 15:26'!
isAbstractClassM1

"A class ?Class is an abstract class if it has an abstract method in its class chain that is not overridden"

?Class isAbstractClass if 
	?Method isMethodInClassChainOfClass: ?Class,?Method isAbstractMethod,not(and(?M2 isMethodInClassChainOfClass: ?Class,?M2 methodOverridesMethod: ?Method))! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isAbstractMethodInClassXM2

"The class ?Class implements an abstract method ?Method "

?Method isAbstractMethodInClass: ?Class if 
	?Method isMethodInClass: ?Class,?Method isAbstractMethod! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isAbstractMethodM1

"Is method ?Method an abstract method?"

?Method isAbstractMethod if 
	?Method isMethod,[?Method isAbstract]! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isCancelledMethodM1

"Is method ?Method cancelled, meaning that it should not be implemented?"

?Method isCancelledMethod if 
	?Method isMethod,[?Method isCancelled]! !

!BasicReasoning methodsFor: 'classes' stamp: 'sp 24/10/11 15:26'!
isClassAboveXM2

"The class ?Super is a superclass in the hierarchy above ?Class, but excluding the class ?Class itself."

?Super isClassAbove: ?Class if 
	?Class classInheritsFrom: ?Super! !

!BasicReasoning methodsFor: 'classes' stamp: 'sp 24/10/11 15:26'!
isClassBelowXM2

"The class ?Class is a subclass in the hierarchy below ?Super, but excluding the class ?Super itself."

?Class isClassBelow: ?Super if 
	?Class classInheritsFrom: ?Super! !

!BasicReasoning methodsFor: 'classes' stamp: 'sp 24/10/11 15:26'!
isClassInHierarchyOfXM2

"The class ?Sub is a class in the hierarchy of class ?Root (including that rootclass)
Case 1: Both ?Sub and ?Root are bound"

+?Sub isClassInHierarchyOf: +?Root if 
	?Sub isClass,?Root isClass,[MLI current is: ?Sub below: ?Root].
"The class ?Sub is a class in the hierarchy of class ?Root (including that rootclass)
Case 2: ?Sub is bound; ?Root is not bound"

+?Sub isClassInHierarchyOf: -?Root if 
	?Sub isClass,[MLI current allSuperclassesOf: ?Sub] contains: ?Root.
"The class ?Sub is a class in the hierarchy of class ?Root (including that rootclass)
Case 3: ?Sub is not bound; ?Root is  bound"

-?Sub isClassInHierarchyOf: +?Root if 
	?Root isClass,[MLI current allSubclassesOf: ?Root] contains: ?Sub.
"The class ?Sub is a class in the hierarchy of class ?Root (including that rootclass)
Case 4: ?Sub and ?Root are not bound"

-?Sub isClassInHierarchyOf: -?Root if 
	?Root isClass,?Sub isClassInHierarchyOf: ?Root! !

!BasicReasoning methodsFor: 'naming' stamp: 'sp 24/10/11 15:26'!
isClassInPackageNamedXM2

"Class ?Class is defined in a package named ?PackageName
CASE 1: ?PackageName is bound"

?Class isClassInPackageNamed: +?PackageName if 
	?Package isPackageWithName: ?PackageName,?Class isClassInPackage: ?Package.
"Class ?Class is defined in a package named ?PackageName
CASE 2: ?PackageName is not bound"

?Class isClassInPackageNamed: -?PackageName if 
	?Class isClassInPackage: ?Package,?Package isPackageWithName: ?PackageName! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isClassMethodM1

"Is the ?Method a class method?"

?Method isClassMethod if 
	?Method isMethodInClass: ?Class,?Class isMetaClass! !

!BasicReasoning methodsFor: 'naming' stamp: 'AndyKellens 11/22/2011 07:59'!
isClassWithNameXM2

"The class ?Class is named ?ClassName.
CASE 1:?ClassName is bound"

?Class isClassWithName: +?ClassName if 
	[MLI current classesFor: ?ClassName asString] contains: ?Class.

"The class ?Class is named ?ClassName.
CASE 2:?ClassName is not bound"

?Class isClassWithName: -?ClassName if 
	?Class isClass,?ClassName equals: [MLI current classNameOf: ?Class]! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
isLastMessageInCascadeXM2

"Is ?Selector the last message that is sent in a ?Cascade?"

?Selector isLastMessageInCascade: RBCascadeNode(?MsgNodes) if 
	?LastMsgNode lastMemberOf: ?MsgNodes,RBMessageNode(?,?Selector,?) equals: ?LastMsgNode! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isMethodInClassChainOfClassXM2

"Returns all methods ?Method implemented in the class chain of ?Class (either by ?Class itself or by any of its superclasses)

Case 1: ?Class implements ?Method itself"

?Method isMethodInClassChainOfClass: ?Class if 
	?Method isMethodInClass: ?Class.
"Returns all methods ?Method implemented in the class chain of ?Class (either by ?Class itself or by any of its superclasses)

Case 2: A superclass of ?Class implements ?Method"

?Method isMethodInClassChainOfClass: ?Class if 
	?Superclass isClassAbove: ?Class,?Method isMethodInClass: ?Superclass! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isMethodInClassXinProtocolXM3

"Is ?Method of class ?Class classified in protocol ?Protocol?
Case 1: ?Method is bound"

+?Method isMethodInClass: ?Class inProtocol: ?Protocol if 
	?Method isMethodWithName: ?Selector inClass: ?Class,?Selector isSelectorOfClass: ?Class inProtocol: ?Protocol.
"Is ?Method of class ?Class classified in protocol ?Protocol?
Case 2: ?Method is bound"

-?Method isMethodInClass: ?Class inProtocol: ?Protocol if 
	?Selector isSelectorOfClass: ?Class inProtocol: ?Protocol,?Method isMethodWithName: ?Selector inClass: ?Class! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isMethodInProtocolXM2

"Is ?Method classified in protocol ?Protocol?"

?Method isMethodInProtocol: ?Protocol if 
	?Method isMethodInClass: ? inProtocol: ?Protocol! !

!BasicReasoning methodsFor: 'naming' stamp: 'sp 24/10/11 15:26'!
isMethodWithNameXM2

"Retrieve the name ?Selector of a method ?Method"

?Method isMethodWithName: ?Selector if 
	?Method isMethod,[MLI current methodNameOf: ?Method] equals: ?Selector! !

!BasicReasoning methodsFor: 'naming' stamp: 'sp 24/10/11 15:26'!
isMethodWithNameXinClassXM3

"defines the name (selector) of a method in a class."

?Method isMethodWithName: ?Selector inClass: ?Class if 
	?Method isMethodInClass: ?Class,?Method isMethodWithName: ?Selector! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isOverriddenSelectorInClassXM2

"Is ?Selector an overridden selector in class ?Class?"

?Selector isOverriddenSelectorInClass: ?Class if 
	?Class classImplementsSelector: ?Selector,?Super isSuperclassOf: ?Class,?Super classUnderstandsSelector: ?Selector! !

!BasicReasoning methodsFor: 'naming' stamp: 'sp 24/10/11 15:26'!
isPackageWithNameXM2

"Retrieves the ?Name of a ?Package"

?Package isPackageWithName: ?Name if 
	?Package isPackage,?Name equals: [MLI forSmalltalk nameForPackage: ?Package]! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isPrimitiveMethodM1

"Is the ?Method a primitive method"

?Method isPrimitiveMethod if 
	?Method isMethod,[?Method isPrimitive]! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isSelectorM1

"Is ?Selector an existing selector in the system?"

?Selector isSelector if 
	? classImplementsSelector: ?Selector,!!! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isSelectorOfClassXinProtocolXM3

"isSelectorOfClass: verifies whether ?Selector is the name of a
	 method in the protocol with name ?Protocol of class ?Class, or works in a more generative way
	 if one or more of the variables are uninstantiated.

Case 1: ?Selector is unbound"

-?Selector isSelectorOfClass: ?Class inProtocol: ?Protocol if 
	?Protocol isProtocolInClass: ?Class,[MLI current selectorsOfClass: ?Class inProtocol: ?Protocol] contains: ?Selector.
"isSelectorOfClass: verifies whether ?Selector is the name of a
	 method in the protocol with name ?Protocol of class ?Class, or works in a more generative way
	 if one or more of the variables are uninstantiated.

Case 2: ?Selector is bound"

+?Selector isSelectorOfClass: ?Class inProtocol: ?Protocol if 
	?Protocol isProtocolInClass: ?Class,[MLI current isSelector: ?Selector ofClass: ?Class inProtocol: ?Protocol]! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
isSourceCodeOfMethodXM2

"Retrieve the source code of a ?Method"

?Source isSourceCodeOfMethod: ?Method if 
	?Method isMethod,?Source equals: [?Method sourceCode]! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
isStatementOfMethodXM2

"Is ?Statement a statement in the method ?Method"

?Statement isStatementOfMethod: ?Method if 
	?Statements areStatementsOfMethod: ?Method,?Statements contains: ?Statement! !

!BasicReasoning methodsFor: 'classes' stamp: 'sp 24/10/11 15:26'!
isSubclassOfXM2

"The class ?SubClass is a direct subclass of ?SuperClass"

?SubClass isSubclassOf: ?SuperClass if 
	?SuperClass isSuperclassOf: ?SubClass! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
isTemporaryVariableOfMethodXM2

"Retrieve the temporary variables of a ?Method"

?Variable isTemporaryVariableOfMethod: ?Method if 
	RBMethodNode(?,?,?body) isParseTreeOfMethod: ?Method,[?body allTemporaryVariables] contains: ?Variable! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodAccessesInstanceVariableXinClassXM3

"?Method of ?Class either reads or writes ?Variable
CASE 1: the method assigns"

?Method methodAccessesInstanceVariable: ?Variable inClass: ?class if 
	?Method methodAssignsInstanceVariable: ?Variable inClass: ?class.
"?Method of ?Class either reads or writes ?Variable
CASE 2: the method reads"

?Method methodAccessesInstanceVariable: ?Variable inClass: ?class if 
	?Method methodReadsInstanceVariable: ?Variable inClass: ?class! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodAssignsInstanceVariableXinClassXM3

"?Does ?Method of ?Class assign to variable ?Variable?"

?Method methodAssignsInstanceVariable: ?Variable inClass: ?Class if 
	?Method isMethodInClass: ?Class,?Method methodAssignsVariable: ?Variable expression: ?,?Variable isInstanceVariableInClassChainOf: ?Class! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodAssignsVariableXexpressionXM3

"Does the method ?Method contain an assignment to ?Variable  (temp or instance variable) with ?Expression as the rhs argument?"

?Method methodAssignsVariable: ?Variable expression: ?Expression if 
	?Method isMethod,?Tree isParseTreeOfMethod: ?Method,RBAssignmentNode(RBVariableNode(?Variable),?Expression) isStChildOf: ?Tree! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodCallsMethodXM2

"Does ?Method1 potentially call ?Method2?
This is only a coarse-grained over-estimation: all methods with the correct selector are looked for

Case 1: ?Method2 is not bound; find all methods matching a selector invoked in ?Method1"

?Method1 methodCallsMethod: -?Method2 if 
	?Method1 methodSendsSelector: ?Selector,?Method2 isMethodWithName: ?Selector.
"Does ?Method1 potentially call ?Method2?
This is only a coarse-grained over-estimation: all methods with the correct selector are looked for

Case 2: ?Method2 is bound; Check if ?Method1 calls the selector implemented by ?Method2"

?Method1 methodCallsMethod: +?Method2 if 
	?Method2 isMethodWithName: ?Selector,?Method1 methodSendsSelector: ?Selector! !

!BasicReasoning methodsFor: 'methods' stamp: 'sp 24/10/11 15:26'!
methodOverridesMethodXM2

"Does ?Method1 override ?Method2?

Case 1: Both arguments are bound"

+?Method1 methodOverridesMethod: +?Method2 if 
	?Method1 isMethodWithName: ?Selector inClass: ?Class1,?Method2 isMethodWithName: ?Selector inClass: ?Class2,?Class1 isClassBelow: ?Class2.
"Does ?Method1 override ?Method2?

Case 2: ?Method1 is bound; ?Method2 is not bound"

+?Method1 methodOverridesMethod: -?Method2 if 
	?Method1 isMethodWithName: ?Selector inClass: ?Class1,?Class1 isClassBelow: ?Class2,?Method2 isMethodWithName: ?Selector inClass: ?Class2.
"Does ?Method1 override ?Method2?

Case 3: ?Method1 is not bound; ?Method2 is bound"

-?Method1 methodOverridesMethod: +?Method2 if 
	?Method2 isMethodWithName: ?Selector inClass: ?Class2,?Class1 isClassBelow: ?Class2,?Method1 isMethodWithName: ?Selector inClass: ?Class1! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodReadsInstanceVariableXinClassXM3

"Does ?Method read instance variable ?Variable of class ?Class"

?Method methodReadsInstanceVariable: ?Variable inClass: ?class if 
	?Method isMethodWithName: ?Selector inClass: ?class,?Variable isInstanceVariableInClassChainOf: ?Class,[MLI forSmalltalk
	selectorsOfClass: ?Class
	whichRead: ?Variable asString] contains: ?Selector! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodReferencesClassXM2

"Does method ?Method refer to class ?Class?"

?Method methodReferencesClass: ?Class if 
	?Method isMethod,[?Method classesReferenced] contains: ?Class! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodReturnsBooleanXM2

"Does this ?Method return a ?Boolean?"

?Method methodReturnsBoolean: ?Boolean if 
	?Method methodReturnsExpression: RBLiteralValueNode(?Boolean),[?Boolean isKindOf: Boolean]! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodReturnsExpressionXM2

"Does this ?Method have a return expression ?Expression"

?Method methodReturnsExpression: ?Expression if 
	?Method isMethod,?Tree isParseTreeOfMethod: ?Method,RBReturnNode(?Expression) isStChildOf: ?Tree! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodSendsSelectorXM2

"Does the ?Method send selector ?Selector?"

?Method methodSendsSelector: ?Selector if 
	?Method isMethod,?tree isParseTreeOfMethod: ?Method,RBMessageNode(?,?Selector,?) isStChildOf: ?tree! !

!BasicReasoning methodsFor: 'method body' stamp: 'sp 24/10/11 15:26'!
methodWithUniqueStatementXM2

"Does this ?Method have a unique ?Statement?"

?Method methodWithUniqueStatement: ?Statement if 
	<?Statement> areStatementsOfMethod: ?Method! !

SmalltalkReasoning subclass: #Reification
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LiCoR-Predicates'!

!Reification commentStamp: '<historical>' prior: 0!
The Reification layer reifies the basic Smalltalk entities.

It is split up in PackagingReification and SmalltalkReification.
The reason for doing so is that packaging is essentially an environment issue
(there are no real language primitives that deal with packaging), whereas the
Smalltalk reification reifies the Smalltalk primitive language constructs.!

Reification subclass: #PackagingReification
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LiCoR-Predicates'!

!PackagingReification commentStamp: '<historical>' prior: 0!
PackagingReification reifies the entities that deal with packaging. Essentially this is environment-specific and not language-specific, as there are no real Smalltalk language primitives that deal with packaging.!

!PackagingReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isClassCategoryM1

"Is ?Category a class category
CASE 1: ?Category is not bound"

-?Category isClassCategory if 
	member(?Category,[MLI forSmalltalk allCategoryEntities]).
"Is ?Category a class category
CASE 2: ?Category is  bound"

+?Category isClassCategory if 
	[MLI forSmalltalk isCategory: ?Category]! !

!PackagingReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isClassInCategoryXM2

"Belongs ?Class to category ?Category"

?Class isClassInCategory: ?Category if 
	?Class isClass,?Category equals: [MLI forSmalltalk categoryForClass: ?Class]! !

!PackagingReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isClassInPackageXM2

"Predicate that verifies whether ?Class belongs to ?Package
Case 1: Both ?Class and ?Package are bound"

+?Class isClassInPackage: +?Package if 
	?Class isClass,?Package isPackage,[MLI forSmalltalk isClass: ?Class inPackage: ?Package].
"Predicate that verifies whether ?Class belongs to ?Package
Case 2:  ?Class is unbound;  ?Package is bound"

-?Class isClassInPackage: +?Package if 
	?Package isPackage,[MLI forSmalltalk classesInPackage: ?Package] contains: ?Class.
"Predicate that verifies whether ?Class belongs to ?Package
Case 3:  ?MetaClass is a meta class (bound);  ?Package is bound"

+?MetaClass isClassInPackage: +?Package if 
	?MetaClass isMetaClass,?Package isPackage,?MetaClass isMetaClassOf: ?Class,[MLI forSmalltalk classesInPackage: ?Package] contains: ?Class.
"Predicate that verifies whether ?Class belongs to ?Package
Case 4:  ?Package is unbound"

?Class isClassInPackage: -?Package if 
	?Class isClass,?Package equals: [MLI forSmalltalk packageForClass: ?Class]! !

!PackagingReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isMethodInPackageXM2

"Does ?Method belong to ?Package
Case 1: ?Package is bound"

?Method isMethodInPackage: +?Package if 
	?Package isPackage,[MLI forSmalltalk methodsInPackage: ?Package] contains: ?Method.
"Does ?Method belong to ?Package
Case 2: ?Package is not bound"

?Method isMethodInPackage: -?Package if 
	?Method isMethod,?Package equals: [MLI forSmalltalk packageForMethod: ?Method]! !

!PackagingReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isPackageM1

"Checks whether ?Package is a package, or retrieves all packages
Case 1: ?Package is bound"

+?Package isPackage if 
	[MLI forSmalltalk isPackage: ?Package].
"Checks whether ?Package is a package, or retrieves all packages
Case 1: ?Package is not bound"

-?Package isPackage if 
	[MLI forSmalltalk allPackageEntities] contains: ?Package! !

!PackagingReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isProtocolInClassXM2

"isProtocolInClass/2 implements the 'is in' relation between protocols and classes. You can use it with any of the variables bound or unbound. 	As 	such, you can verify a relation, extract all relations, extract all protocols from a class or find the class in which a protocol is defined

CASE 1: ?Protocol unbound"

-?Protocol isProtocolInClass: ?Class if 
	?Class isClass,[MLI forSmalltalk allProtocolsOfClass: ?Class] contains: ?Protocol.
"isProtocolInClass/2 implements the 'is in' relation between protocols and classes. You can use it with any of the variables bound or unbound. 	As 	such, you can verify a relation, extract all relations, extract all protocols from a class or find the class in which a protocol is defined

CASE 2: ?Protocol bound"

+?Protocol isProtocolInClass: ?Class if 
	?Class isClass,[MLI forSmalltalk isProtocol: ?Protocol ofClass: ?Class]! !

!PackagingReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isProtocolM1

"Is ?Protocol a method protocol?"

?Protocol isProtocol if 
	member(?Protocol,[MLI forSmalltalk allProtocolEntities])! !

!Reification methodsFor: 'method attributes' stamp: 'sp 24/10/11 15:26'!
methodWithPragmaNamedM2

methodWithPragmaNamed(-?method,+?name) if 
	class(?class),member(?method,[MLI forSmalltalk
	methodsInClass: ?class
	withPragmaName: ?name asSymbol]).
methodWithPragmaNamed(+?method,+?name) if 
	methodInClass(?method,?class),member(?method,[MLI forSmalltalk
	methodsInClass: ?class
	withPragmaName: ?name asSymbol])! !

Reification subclass: #SmalltalkReification
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LiCoR-Predicates'!

!SmalltalkReification commentStamp: '<historical>' prior: 0!
SmalltalkReification reifies the real Smalltalk lanuage entities. Is not supposed to contain any environment-specific stuff.!

!SmalltalkReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isBaseClassM1

"This predicate verifies whether ?Class is a base class

CASE 1: the variable ?Class is bound"

+?class isBaseClass if 
	[MLI forSmalltalk isBaseClass: ?class].
"This predicate verifies whether ?Class is a base class 

CASE 2: the variable ?Class is not bound"

-?Class isBaseClass if 
	[MLI forSmalltalk allBaseClassEntities] contains: ?Class! !

!SmalltalkReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isClassM1

"This predicate verifies whether ?Class is a base or meta class

CASE 1: the variable ?Class is bound"

+?Class isClass if 
	[MLI forSmalltalk isClass: ?Class].
"This predicate verifies whether ?Class is a base or meta class

CASE 2: the variable ?Class is not bound"

-?Class isClass if 
	[MLI forSmalltalk allClassEntities] contains: ?Class! !

!SmalltalkReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isClassMethodInClassXM2

"?Method isClassMethodInClass: ?Class verifies whether a ?Class contains
	 a given class ?Method or returns the appropriate value if one (or more) of the arguments
	 is variable.

CASE 1: ?Method is known"

+?Method isClassMethodInClass: ?Class if 
	?Class isClass,[MLI forSmalltalk isClassMethod: ?Method inClass: ?Class].
"CASE 2: ?Method is unknown"

-?Method isClassMethodInClass: ?Class if 
	?Class isClass,[MLI forSmalltalk classMethodsInClass: ?Class] contains: ?Method! !

!SmalltalkReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isClassVariableInClassXM2

"?ClassVariable isClassVariableInClass: ?Class verifies whether a ?Class contains
	 a given ?ClassVariable or returns the appropriate values if one (or more) of the arguments
	 is variable.

CASE 1: +?ClassVariable isClassVariableInClass: ?Class"

+?ClassVariable isClassVariableInClass: ?Class if 
	?Class isClass,[MLI forSmalltalk isClassVariable: ?ClassVariable inClass: ?Class].
"CASE 2: -?ClassVariable isClassVariableInClass: ?Class"

-?ClassVariable isClassVariableInClass: ?Class if 
	?Class isClass,[MLI forSmalltalk allClassVariableEntitiesInClass: ?Class] contains: ?ClassVariable! !

!SmalltalkReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isClassVariableM1

"+?ClassVar isClassVariableverifies whether ?ClassVar is a class variable of some existing class
	 or returns the appropriate value if ?ClassVar is variable.

CASE 1:+?ClassVar isClassVariable checks whether ?ClassVar is an existing class variable"

+?ClassVar isClassVariable if 
	[MLI forSmalltalk isClassVariable: ?ClassVar].
"CASE 2: -?ClassVar isClassVariable returns an existing class variable"

-?ClassVar isClassVariable if 
	[MLI forSmalltalk allClassVariableEntities] contains: ?ClassVar! !

!SmalltalkReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isInstanceVariableInClassChainOfXM2

"?InstanceVariable isInstanceVariableInClassChainOf: ?Class verifies whether a ?Class or
	 one of its (direct or indirect) superclasses declares a given ?InstanceVariable. If ?Class is
	 variable, the class that declares the ?InstanceVariable as well as all of its subclasses are
	 returned. If ?InstanceVariable is variable, all instance variables that are declared in ?Class
	 or one of its superclasses are returned. If both are variable we have the combination of
	 these two situations and we produce all appropriate solutions.

CASE 1: +?InstanceVariable isInstanceVariableInClassChainOf: ?Class"

+?InstanceVariable isInstanceVariableInClassChainOf: ?Class if 
	?Class isClass,[MLI forSmalltalk isInstanceVariable: ?InstanceVariable inClassChain: ?Class].
"CASE 2: -?InstanceVariable isInstanceVariableInClassChainOf: ?Class"

-?InstanceVariable isInstanceVariableInClassChainOf: ?Class if 
	?Class isClass,[MLI forSmalltalk allInstanceVariableEntitiesInClassChain: ?Class] contains: ?InstanceVariable! !

!SmalltalkReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isInstanceVariableInClassXM2

"+?InstanceVariable isInstanceVariableInClass: ?Class verifies whether a ?Class contains
	 a given ?InstanceVariable or returns the appropriate values if one (or more) of the arguments
	 is variable.

CASE 1: +?InstanceVariable isInstanceVariableInClass: +-?Class"

+?InstanceVariable isInstanceVariableInClass: ?Class if 
	?Class isClass,[MLI forSmalltalk isInstanceVariable: ?InstanceVariable inClass: ?Class].
"CASE 2: -?InstanceVariable isInstanceVariableInClass: +-?Class"

-?InstanceVariable isInstanceVariableInClass: ?Class if 
	?Class isClass,[MLI forSmalltalk allInstanceVariableEntitiesInClass: ?Class] contains: ?InstanceVariable! !

!SmalltalkReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isInstanceVariableInMetaClassChainOfXM2

"?InstanceVariable isInstanceVariableInMetaClassChainOf: ?Class verifies whether a ?Class or
	 one of its (direct or indirect) superclasses declares a given ?InstanceVariable. If ?Class is
	 variable, the class that declares the ?InstanceVariable as well as all of its subclasses are
	 returned. If ?InstanceVariable is variable, all instance variables that are declared in ?Class
	 or one of its superclasses are returned. If both are variable we have the combination of
	 these two situations and we produce all appropriate solutions."

?InstanceVariable isInstanceVariableInMetaClassChainOf: ?Class if 
	?Class isMetaClass,?InstanceVariable isInstanceVariableInClassChainOf: ?Class! !

!SmalltalkReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isInstanceVariableInMetaClassXM2

"?InstanceVariable isInstanceVariableInMetaClass: ?Class verifies whether a metaclass ?Class contains
	 a given ?InstanceVariable or returns the appropriate values if one (or more) of the arguments
	 is variable."

?InstanceVariable isInstanceVariableInMetaClass: ?Class if 
	?Class isMetaClass,?InstanceVariable isInstanceVariableInClass: ?Class! !

!SmalltalkReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isInstanceVariableM1

"+?InstVar isInstanceVariable verifies whether ?InstVar is an instance variable of some existing Smalltalk class
	 or returns the appropriate value if ?InstVar is variable.

CASE 1: +?InstVar isInstanceVariable checks whether ?InstVar is an existing instance variable"

+?InstVar isInstanceVariable if 
	[MLI forSmalltalk isInstanceVariable: ?InstVar].
"CASE 2: -?InstVar isInstanceVariable returns an existing instance variable"

-?InstVar isInstanceVariable if 
	[MLI forSmalltalk allInstanceVariableEntities] contains: ?InstVar! !

!SmalltalkReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isMetaClassM1

"This predicate verifies whether ?Class is a meta class

CASE 1: the variable ?Class is bound"

+?Class isMetaClass if 
	[MLI forSmalltalk isMetaClass: ?Class].
"This predicate verifies whether ?Class is a meta class

CASE 2: the variable ?Class is not bound"

-?Class isMetaClass if 
	[MLI forSmalltalk allMetaClassEntities] contains: ?Class! !

!SmalltalkReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isMetaClassOfXM2

"?MetaClass isMetaClassOf: ?Class verifies whether ?MetaClass is the Smalltalk metaclass
	 of ?Class. It works for ?MetaClass and ?Class bound or variable"

?MetaClass isMetaClassOf: ?Class if 
	?Class isClass,?MetaClass equals: [MLI forSmalltalk metaClassOf: ?Class]! !

!SmalltalkReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isMethodInClassXM2

"+?Method isMethodInClass: ?Class  verifies whether a ?Class contains
	 a given ?Method or returns the appropriate value if one (or more) of the arguments
	 is variable.

CASE 1: ?Method is known. Assumption: a method is in ONE class."

+?Method isMethodInClass: ?Class if 
	?Method isMethod,?Class equals: [MLI forSmalltalk classOfMethod: ?Method].
"CASE 2: ?Method is unknown"

-?Method isMethodInClass: ?Class if 
	?Class isClass,[MLI forSmalltalk methodsInClass: ?Class] contains: ?Method! !

!SmalltalkReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isMethodM1

"+-?Method isMethod  verifies whether ?Method is a Smalltalk method. If ?Method is a variable
	 it will be instantiated with a method. If ?Method is bound it should be an
	 existing method in the current Smalltalk image.

CASE 1: +Method isMethod checks whether ?method is an existing method"

+?Method isMethod if 
	[MLI forSmalltalk isMethod: ?Method].
"CASE 2: -Method isMethod returns an existing method"

-?Method isMethod if 
	[MLI forSmalltalk allMethodEntities] contains: ?Method! !

!SmalltalkReification methodsFor: 'method attributes' stamp: 'sp 24/10/11 15:26'!
isParseTreeOfMethodXM2

?tree isParseTreeOfMethod: ?method if 
	?method isMethod,?tree equals: [?method rbParseTree],[?tree notNil]! !

!SmalltalkReification methodsFor: 'entities' stamp: 'sp 24/10/11 15:26'!
isRootClassM1

"isRootClass verifies whether ?Class is a root class, i.e., whether it has
	 no superclass. If ?Class is variable it will be instantiated with such a class.
	 If ?Class is bound it should be an existing root class in the current system.

CASE 1: rootClass(+?Class) checks whether ?Class is an existing root class"

+?Class isRootClass if 
	[MLI forSmalltalk isRootClass: ?Class].
"CASE 2: isRootClass returns an existing root class"

-?Class isRootClass if 
	[MLI forSmalltalk allRootClassEntities] contains: ?Class! !

!SmalltalkReification methodsFor: 'method attributes' stamp: 'sp 24/10/11 15:26'!
isSourceCodeOfMethodXM2

"Retrieve the ?Source of a ?Method"

?Source isSourceCodeOfMethod: ?Method if 
	?Method isMethod,?Source equals: [?Method sourceCode]! !

!SmalltalkReification methodsFor: 'parsetree traversal' stamp: 'sp 24/10/11 15:26'!
isStChildOfXM2

?aTerm isStChildOf: ?aTerm if 
	not([?aTerm isKindOf: OrderedCollection]).
?aTerm isStChildOf: ?afunctor@(?args) if 
	?args contains: ?child,?aTerm isStChildOf: ?child! !

!SmalltalkReification methodsFor: 'binary relations' stamp: 'sp 24/10/11 15:26'!
isSuperclassOfXM2

"?SuperClass isSuperclassOf: ?SubClass verifies whether ?SuperClass is a super class of
	 ?SubClass in the Smalltalk class hierachy. If one (or both) arguments are unbound,
	 the appropriate values will be produced for them.

Implementation note: In the implementation below some of the cases could have been
	 taken together in a single case, but for performance reasons we preferred to handle all
	 cases separately. This is a frequently used predicate that should be implemented as
	 efficient as possible.

CASE 1: +?SuperClass isSuperclassOf: +?SubClass"

+?SuperClass isSuperclassOf: +?SubClass if 
	?SuperClass isClass,?SubClass isClass,[MLI forSmalltalk is: ?SuperClass superclassOf: ?SubClass].
"CASE 2: -?SuperClass isSuperclassOf: +?SubClass"

-?SuperClass isSuperclassOf: +?SubClass if 
	?SubClass isClass,?SuperClass equals: [MLI forSmalltalk hasSuperclass: ?SubClass],[?SuperClass isNil not].
"CASE 3 & 4: +-?SuperClass isSuperclassOf: -?SubClass

?SuperClass may be either variable or not but it must be a class"

?SuperClass isSuperclassOf: -?SubClass if 
	?SuperClass isClass,[MLI forSmalltalk subclassesOf: ?SuperClass] contains: ?SubClass! !

!SmalltalkReification methodsFor: 'method attributes' stamp: 'sp 24/10/11 15:26'!
methodContainsSourceCommentXM2

"Is ?Comment a comment in the source code of method ?Method"

?Method methodContainsSourceComment: ?Comment if 
	?Method isMethod,[MLI forSmalltalk sourceCommentsForMethod: ?Method] contains: ?Comment! !

!SmalltalkReification methodsFor: 'method attributes' stamp: 'sp 24/10/11 15:26'!
methodHasAnnotationXwithValuesXM3

"This predicate retrieves all ?Annotations of a ?Method, together with the list of ?Values of that annotation"

?Method methodHasAnnotation: ?Annotation withValues: ?Values if 
	?Method isMethod,[MLI forSmalltalk annotationsOfMethod: ?Method] contains: ?association,?Annotation equals: [?association key],?Values equals: [?association value]! !
